from typing import List, Union, Optional, Type

from loguru import logger

from iot.context import Context
from iot.models import OnStartUpHook, OnTearDownHook, OnModelPreprocessHook, OnModelPostprocessHook, \
    FunctionDeviceModel, UniverseNoticeModel
from iot.semantics import SemanticsGroup


class HooksGroup:
    __on_start_up_hooks: List['OnStartUpHook'] = []
    __on_tear_down_hooks: List['OnTearDownHook'] = []
    __on_model_preprocess_hooks: List['OnModelPreprocessHook'] = []
    __on_model_postprocess_hooks: List['OnModelPostprocessHook']= []

    @classmethod
    def on_start_up(cls, on_start_up_hook: OnStartUpHook):
        cls.__on_start_up_hooks.append(on_start_up_hook)
        logger.info("已加载钩子函数 " + on_start_up_hook.__str__())

    @classmethod
    def on_tear_down(cls, on_tear_down_hook: OnTearDownHook):
        cls.__on_tear_down_hooks.append(on_tear_down_hook)
        logger.info("已加载钩子函数 " + on_tear_down_hook.__str__())

    @classmethod
    def on_model_preprocess(cls, on_model_preprocess_hook: OnModelPreprocessHook):
        cls.__on_model_preprocess_hooks.append(on_model_preprocess_hook)
        logger.info("已加载钩子函数 " + on_model_preprocess_hook.__str__())

    @classmethod
    def on_model_postprocess(cls, on_model_postprocess_hook: OnModelPostprocessHook):
        cls.__on_model_postprocess_hooks.append(on_model_postprocess_hook)
        logger.info("已加载钩子函数 " + on_model_postprocess_hook.__str__())

    @classmethod
    def execute_start_up(cls, context: Type[Context], semantics_group: Type[SemanticsGroup]):
        for hook in cls.__on_start_up_hooks:
            hook(context=context, semantics_group=semantics_group)

    @classmethod
    def execute_tear_down(cls, context: Type[Context], semantics_group: Type[SemanticsGroup]):
        for hook in cls.__on_tear_down_hooks:
            hook(context=context, semantics_group=semantics_group)

    @classmethod
    def execute_model_preprocess(cls, context: Type[Context], model: Union[FunctionDeviceModel, UniverseNoticeModel]):
        for hook in cls.__on_model_preprocess_hooks:
            hook(context, model)

    @classmethod
    def execute_model_postprocessor(cls, context: Type[Context], model: Optional[Union[FunctionDeviceModel, UniverseNoticeModel]]):
        for hook in cls.__on_model_postprocess_hooks:
            hook(context, model)


import re
from typing import Dict, Tuple, Union, ClassVar, Optional

from loguru import logger

from iot.models import SemanticsModel, SemanticsFromEnum, SemanticsRedirectEnum, FunctionDeviceModel, UniverseNoticeModel


class SemanticsGroup:
    # 来自用户的语义模型，key为正则表达式，value为语义模型
    __inner_group_from_users: ClassVar[Dict[str, "SemanticsModel"]] = {}
    # 来自设备的语义模型，key为topic，value为语义模型
    __inner_group_from_devices: ClassVar[Dict[str, "SemanticsModel"]] = {}

    @classmethod
    def add_model(cls, model: type):
        model = model()
        if model.frm == SemanticsFromEnum.USER:
            cls.__inner_group_from_users[model.regex] = model
            logger.info("已加载用户语义模型：" + model.code)
        else:
            cls.__inner_group_from_devices[model.topic] = model
            logger.info("已加载设备语义模型：" + model.code)

    @classmethod
    def get_model(cls, key: str) -> Optional["SemanticsModel"]:
        for k in cls.__inner_group_from_users.keys():
            if k == key or re.match(k, key):
                return cls.__inner_group_from_users[k]
        for k in cls.__inner_group_from_devices.keys():
            if k == key:
                return cls.__inner_group_from_devices['k']
        return None

    @classmethod
    def process_usr_command(cls, function_device_model: FunctionDeviceModel) -> Tuple[Optional[FunctionDeviceModel],
                                                                                      SemanticsRedirectEnum]:
        if function_device_model.is_raw:
            for pattern in cls.__inner_group_from_users.keys():
                if res := re.match(pattern, function_device_model.data):
                    model = cls.__inner_group_from_users[pattern]
                    topic = model.topic
                    logger.info(f"{function_device_model.data}hits model: {model.code}")
                    args = []
                    for i in range(model.regex_num):
                        args.append(res.group(i))
                    called: Union[FunctionDeviceModel, UniverseNoticeModel]
                    try:
                        logger.info(f"{model.func} called，args：" + " ".join(args))
                        called = model.func(*args)
                        called.topic = topic
                        logger.success(f"{model.func}called with success：" + called.__str__())
                    except Exception as e:
                        logger.error(f"{model.func}called in ERROR：{e}")
                        return None, SemanticsRedirectEnum.MISS
                    if model.redirect == SemanticsRedirectEnum.SEMANTICS:
                        logger.info(f"SemanticsModel redirects to SEMANTICS")
                        return cls.process_usr_command(called)
                    return called, model.redirect
        else:
            for pattern, model in cls.__inner_group_from_users.items():
                if model.code == function_device_model.smt_code:
                    logger.info(f"{function_device_model.data}hits model: {model.code}")
                    called: Union[FunctionDeviceModel, UniverseNoticeModel]
                    try:
                        logger.info(f"{model.func} called")
                        called = model.func(function_device_model)
                        logger.success(f"{model.func}called with success：" + called.__str__())
                    except Exception as e:
                        logger.error(f"{model.func}called in ERROR：{e}")
                        return None, SemanticsRedirectEnum.MISS
                    if model.redirect == SemanticsRedirectEnum.SEMANTICS:
                        logger.info(f"SemanticsModel redirects to SEMANTICS")
                        return cls.process_usr_command(called)
                    return called, model.redirect
        logger.warning(f"{function_device_model.data} cannot hit any model")
        return None, SemanticsRedirectEnum.MISS

    @classmethod
    def process_dvc_command(cls, universe_notice_model: UniverseNoticeModel) -> Tuple[Optional[UniverseNoticeModel],
                                                                                      SemanticsRedirectEnum]:
        if not universe_notice_model.verbose:
            logger.info(f"{universe_notice_model.message} returns directly")
            return universe_notice_model, SemanticsRedirectEnum.ACOUSTICS
        for topic in cls.__inner_group_from_devices.keys():
            if topic == universe_notice_model.topic:
                logger.info(f"{universe_notice_model.data} hits topic {topic}")
                model = cls.__inner_group_from_devices[topic]
                called: Union[FunctionDeviceModel, UniverseNoticeModel]
                try:
                    logger.info(f"{model.func} called")
                    called = model.func(universe_notice_model)
                    logger.success(f"{model.func}called with success：" + called.__str__())
                except Exception as e:
                    logger.error(f"{model.func}called in ERROR：{e}")
                    return None, SemanticsRedirectEnum.MISS
                if model.redirect == SemanticsRedirectEnum.SEMANTICS:
                    logger.info(f"SemanticsModel redirects to SEMANTICS")
                    return cls.process_dvc_command(called)
                return called, model.redirect
        logger.warning(f"{universe_notice_model.data} cannot hit any model")
        return None, SemanticsRedirectEnum.MISS



import time as tm

from iot.context import Context
from iot.factories import TtsProcessorFactory
from iot.integration.baidu import BaiduCloudClient, BaiduApi
from iot.models import TtsProcessor
from iot.utils import get_slash


@TtsProcessorFactory.set
class BuiltinTtsProcessor(TtsProcessor):
    client: BaiduCloudClient

    def __init__(self):
        self.client = BaiduApi.register()

    def tts(self, text: str) -> str:
        local_time = tm.strftime('%Y-%m-%d_%H-%M-%S', tm.localtime())
        path = f'{Context.config["RESOURCE_PATH"]}{get_slash()}txt{get_slash()}{local_time}.txt'
        res = self.client.tts(text=text, filename=path)
        return path



from loguru import logger

from iot.context import Context
from iot.factories import RecorderFactory
from iot.integration.audio import AudioClient
from iot.models import Recorder

import keyboard
import time as tm
import os

from iot.utils import get_slash


@RecorderFactory.set
class BuiltinRecorder(Recorder):

    def awake(self) -> bool:
        logger.info("按下f唤醒智能终端...")
        keyboard.wait("f")
        return True

    def record(self, time: int) -> str:
        data = AudioClient.record(time=time)
        local_time = tm.strftime('%Y-%m-%d_%H-%M-%S', tm.localtime())
        path = f'{Context.config["RESOURCE_PATH"]}{get_slash()}wav{get_slash()}{local_time}.wav'
        AudioClient.save(data=data, path=path)
        return path


@RecorderFactory.set
class PocketsphinxRecorder(Recorder):

    def awake(self) -> bool:
        try:
            from pocketsphinx import LiveSpeech, get_model_path
        except ImportError as ie:
            logger.error("没有安装pocketsphinx，具体解决方案请访问https://github.com/bambocher/pocketsphinx-python " + ie.__str__())
            return False

        model_path = get_model_path()
        speech = LiveSpeech(
            verbose=False,
            sampling_rate=16000,
            buffer_size=2048,
            no_search=False,
            full_utt=False,
            hmm=os.path.join(model_path, 'en-us'),
            lm=os.path.join(model_path, 'en-us.lm.bin'),
            dic=os.path.join(model_path, 'cmudict-en-us.dict')
        )
        for phrase in speech:
            logger.info("Got phrase: " + phrase.__str__())
            phrase = phrase.__str__().split()[0]
            if phrase.__str__() in ['hi', 'hello', 'i', 'ah']:
                return True

    def record(self, time: int) -> str:
        data = AudioClient.record(time=time)
        local_time = tm.strftime('%Y-%m-%d_%H-%M-%S', tm.localtime())
        path = f'{Context.config["RESOURCE_PATH"]}{get_slash()}wav{get_slash()}{local_time}.wav'
        AudioClient.save(data=data, path=path)
        return path



from iot.semantics import SemanticsGroup
from iot.models import SemanticsModel, SemanticsFromEnum, SemanticsRedirectEnum, FunctionDeviceModel, SemanticsFunc


def switch_semantics_model_func(*args) -> FunctionDeviceModel:
    param = args[0]
    if param == "打开开关":
        return FunctionDeviceModel(smt_code="default01",
                                   topic="switch",
                                   is_raw=False,
                                   acoustics="好的，已为您打开开关",
                                   data={
                                       'switch': 'on'
                                   })
    else:
        return FunctionDeviceModel(smt_code="default02",
                                   topic="switch",
                                   is_raw=False,
                                   acoustics="好的，已为您关闭开关",
                                   data={
                                       'switch': 'off'
                                   })


@SemanticsGroup.add_model
class SwitchOnSemanticsModel(SemanticsModel):
    code = "default01"
    frm = SemanticsFromEnum.USER
    topic = 'default/switch'
    regex = "打开开关"
    regex_num = 1
    redirect = SemanticsRedirectEnum.MESSAGE
    func: SemanticsFunc = switch_semantics_model_func


@SemanticsGroup.add_model
class SwitchOffSemanticssModel(SemanticsModel):
    code = "default02"
    frm = SemanticsFromEnum.USER
    topic = 'default/switch'
    regex = "关闭开关"
    regex_num = 1
    redirect = SemanticsRedirectEnum.MESSAGE
    func: SemanticsFunc = switch_semantics_model_func



import base64
import json
from typing import Optional
from urllib.request import urlopen
from urllib.request import Request
from urllib.parse import quote_plus, urlencode
from urllib.error import URLError

import requests
from loguru import logger

from iot.context import Context


class BaiduCloudClient:
    __access_token: str = None

    def __init__(self, api_key: str, secret_key: str, access_token: str = None):
        if access_token:
            self.__access_token = access_token
        else:
            url = f'https://aip.baidubce.com/oauth/2.0/token?' \
                  f'grant_type=client_credentials&client_id={api_key}&client_secret={secret_key}'
            response = requests.get(url)
            if response:
                self.__access_token = response.json()['access_token']
                logger.info("access_token " + self.__access_token)

    def asr(self, path: str) -> Optional[str]:
        speech_data = []
        with open(path, 'rb') as speech_file:
            speech_data = speech_file.read()
        length = len(speech_data)
        if length == 0:
            logger.error("Cannot read anything from " + path)
            return None
        speech = base64.b64encode(speech_data)
        speech = str(speech, 'utf-8')
        params = {'dev_pid': 1537,
                  'format': path[-3:],
                  'rate': 16000,
                  'token': self.__access_token,
                  'cuid': 'fubuki_cuid',
                  'channel': 1,
                  'speech': speech,
                  'len': length
                  }
        post_data = json.dumps(params, sort_keys=False)
        req = Request('http://vop.baidu.com/server_api', post_data.encode('utf-8'))
        req.add_header('Content-Type', 'application/json')
        try:
            f = urlopen(req)
            result_str = f.read()
            result = json.loads(result_str)
            if result['err_no'] != 0:
                logger.error("Fail to process ASR " + result['err_msg'])
                return None
            result_str = ' '.join(result['result'])
        except URLError as e:
            logger.error(f"Fail to acquire ASR result from BaiduCloud " + e.reason)
            result_str = None
        return result_str

    def tts(self, text: str, filename: str,
            per: int = 0, spd: int = 5, pit: int = 5, vol: int = 5) -> bool:
        text = quote_plus(text)
        params = {'tok': self.__access_token,
                  'tex': text,
                  'per': per,
                  'spd': spd,
                  'pit': pit,
                  'vol': vol,
                  'aue': 6, # wav file
                  'cuid': 'fubuki_cuid',
                  'lan': 'zh',
                  'ctp': 1}
        data = urlencode(params)
        req = Request('http://tsn.baidu.com/text2audio', data.encode('utf-8'))
        has_error = False
        try:
            f = urlopen(req)
            result_str = f.read()

            headers = dict((name.lower(), value) for name, value in f.headers.items())

            has_error = ('content-type' not in headers.keys() or headers['content-type'].find('audio/') < 0)
        except URLError as e:
            logger.error(f"Fail to acquire tts result from BaiduCloud " + e.strerror)
            result_str = e.reason
            has_error = True
        if not has_error:
            with open(filename, 'wb') as of:
                of.write(result_str)
        else:
            logger.error(f"Fail to save tts result")
        return not has_error


class BaiduApi:
    client: BaiduCloudClient = None

    @classmethod
    def register(cls) -> BaiduCloudClient:
        if cls.client is None:
            cls.access_token = Context.config['BAIDU_ACCESS_TOKEN']
            cls.api_key = Context.config['BAIDU_API_KEY']
            cls.secret_key = Context.config['BAIDU_SECRET_KEY']
            cls.client = BaiduCloudClient(api_key=cls.api_key, secret_key=cls.secret_key, access_token=cls.access_token)
        return cls.client




from collections import UserDict

from dotenv import dotenv_values


class Config(UserDict):

    def load_config(self):
        self.data = dotenv_values()
        if 'RESOURCE_PATH' not in self.data:
            raise KeyError("Config missing: RESOURCE_PATH")
        if 'DEVICE_REC' not in self.data:
            self['DEVICE_REC'] = 'BuiltinRecorder'
        if 'DEVICE_PLY' not in self.data:
            self['DEVICE_PLY'] = 'BuiltinPlayer'
        if 'ASR_PROCESSOR' not in self.data:
            self['ASR_PROCESSOR'] = 'BuiltinAsrProcessor'
        if 'TTS_PROCESSOR' not in self.data:
            self['TTS_PROCESSOR'] = 'BuiltinTtsProcessor'
        if 'MQTT_HOST' not in self.data:
            self['MQTT_HOST'] = '127.0.0.1'
        if 'MQTT_PORT' not in self.data:
            self['MQTT_PORT'] = '1883'
        if 'TERMINAL_MODE' not in self.data: # 终端模式：0-监听用户 1-监听设备 2-两者都监听
            self['TERMINAL_MODE'] = '2'

    def __getitem__(self, item):
        if not self.data:
            self.load_config()
        return self.data[item]

    def __setitem__(self, key, value):
        if not self.data:
            self.load_config()
        self.data[key] = value



import abc
from enum import Enum
from typing import Optional, Callable, Union, Dict, TYPE_CHECKING

from pydantic import BaseModel


class Factory(abc.ABC):

    @abc.abstractmethod
    def set(cls, target: type):
        raise NotImplementedError

    @abc.abstractmethod
    def get(cls, target_name):
        raise NotImplementedError


class Recorder(abc.ABC):

    @abc.abstractmethod
    def awake(self) -> bool:
        """
        唤醒终端设备，必须是一个阻塞方法
        :return: 返回True表示唤醒，返回False表示终止程序调用终止方法
        """
        raise NotImplementedError

    @abc.abstractmethod
    def record(self, time: int) -> str:
        """
        录音并存储，用于接受用户的指令，建议存储为wav格式
        :param time: 录音时长，默认3秒
        :return: 录音文件的地址
        """
        raise NotImplementedError


class Player(abc.ABC):

    @abc.abstractmethod
    def play(self, path: str) -> None:
        raise NotImplementedError


class AsrProcessor(abc.ABC):

    @abc.abstractmethod
    def asr(self, path: str) -> Optional[str]:
        """
        对给定路径的音频文件进行语音识别
        :param path: 音频文件的路径
        :return: 语音识别结果，即文字信息。返回`None`说明识别失败
        """
        raise NotImplementedError


class TtsProcessor(abc.ABC):

    @abc.abstractmethod
    def tts(self, text: str) -> str:
        """
        对给定文字进行语音合成
        :param text: 给定文字
        :return: 语音合成音频文件的路径，音频文件建议.wav格式
        """
        raise NotImplementedError


class Hook:

    def __call__(self, *args, **kwargs):
        """
        所有钩子函数都是函数
        :param args:
        :param kwargs:
        :return:
        """
        raise NotImplementedError


class OnStartUpHook(Hook):

    def __call__(self, *args, **kwargs):
        if TYPE_CHECKING:
            assert 'context' in kwargs
            assert 'semantics_group' in kwargs


class OnTearDownHook(Hook):

    def __call__(self, *args, **kwargs):
        if TYPE_CHECKING:
            assert 'context' in kwargs
            assert 'semantics_group' in kwargs


class  OnModelPreprocessHook(Hook):

    def __call__(self, *args, **kwargs):
        if TYPE_CHECKING:
            assert 'context' in kwargs
            assert 'function_device_model' in kwargs or 'universe_notice_mode' in kwargs


class OnModelPostprocessHook(Hook):

    def __call__(self, *args, **kwargs):
        if TYPE_CHECKING:
            assert 'context' in kwargs
            assert 'function_device_model' in kwargs or 'universe_notice_mode' in kwargs


class SemanticsFromEnum(Enum):
    USER = 1,
    DEVICE = 2,


class SemanticsRedirectEnum(Enum):
    ACOUSTICS = 1,
    MESSAGE = 2,
    SEMANTICS = 3,
    MISS = 4,


class FunctionDeviceModel(BaseModel):
    """
    功能设备模型，既是用户发送指令的封装，也是语义转换模块处理后的产物。需结合重定向枚举SemanticsRedirectEnum一起用才有明确语义
    """

    # 对应的semantics_model的code，如果is_raw为False则根据这个字段查找对应的semantics_model
    smt_code: Optional[str]

    # 对应的mqtt的topic，如果重定向到MESSAGE则需要给这个topic发送数据
    topic: Optional[str]

    # raw标志位，如果为True则data为str，在语义处理中根据data做正则匹配选择对应的语义模型
    # 如果为False则根据smt_code选择对应的语义模型
    is_raw: bool

    # 执行成功后返回的语音提示，如果重定向到MESSAGE则作为语音提示，
    # 如果是重定向到ACOUSTICS则也可以表示返回内容，
    # 如果重定向到SEMANTICS则没有意义
    acoustics: str

    # 数据，如果重定向到message则该字段作为mqtt发送的payload，
    # 如果是重定向到ACOUSTICS则也可以表示返回内容，
    # 如果重定向到SEMANTICS则再一次根据is_raw字段判断，如果is_raw为True则根据data字段匹配选择对应的语义模型
    data: Union[str, Dict[str, str]]


class UniverseNoticeModel(BaseModel):
    """
    统一推送模型，需结合重定向枚举SemanticsRedirectEnum一起用才有明确语义
    """

    # 对应的semantics_model的code，用户自用
    smt_code: Optional[str]

    # mqtt发送的topic，如果verbose字段为True，则根据topic字段匹配语义模型
    topic: str

    # 发布mqtt消息的设备名称，用户自用
    device: str

    # verbose标志位，如何处理设备推送消息的一句。如果为False则直接将message字段信息作为语音返回，
    # 如果为True，则data不为None，且根据topic查找语义模型
    verbose: bool

    # 返回的语音文本，如果verbose为True或者重定向为ACCOUSTICS，则直接返回该字段作为语音信息
    message: str

    # 数据，如果重定向给MESSAGE，则为FunctionDeviceModel
    data: Optional[Dict[str, str]]


SemanticsFunc = Callable[..., Union[FunctionDeviceModel, UniverseNoticeModel]]


class SemanticsModel(BaseModel):
    """
    语义模型
    """

    # 语义模型标识，自定义，一般用来检索
    code: str

    # 语义来源，用户或者设备
    frm: SemanticsFromEnum

    # topic，用来检索匹配
    topic: str

    # 正则表达式，用来检索匹配
    regex: Optional[str]

    # 正则表达式的group_num
    regex_num: Optional[int]

    # 重定向，结合func的返回结果（模型）决定下一步怎么处理。分为返回语音ACOUSTICS、发布消息MESSAGE、继续语义处理SEMANTICS，
    # 当为MISS时说明没有命中，或者调用func出错
    redirect: SemanticsRedirectEnum

    # 处理语义，若来源为用户则参数为正则匹配结果groups，若为设备则为UniverseNoticeModel.data，
    # 返回元组，第一项是model，第二项是重定向
    func: SemanticsFunc



from iot.terminal import Terminal
from iot.factories import RecorderFactory, PlayerFactory, AsrProcessorFactory, TtsProcessorFactory
from iot.hooks import HooksGroup
from iot.semantics import SemanticsGroup
from iot.models import *


from typing import Optional

from iot.factories import AsrProcessorFactory
from iot.integration.baidu import BaiduCloudClient, BaiduApi
from iot.models import AsrProcessor


@AsrProcessorFactory.set
class BuiltinAsrProcessor(AsrProcessor):
    client: BaiduCloudClient

    def __init__(self):
        self.client = BaiduApi.register()

    def asr(self, path: str) -> Optional[str]:
        result_str = self.client.asr(path)
        return result_str



import json
import os
import platform
import importlib
import sys
from typing import Union, Iterable

import pydantic

from iot.models import FunctionDeviceModel, UniverseNoticeModel


def get_slash():
    if platform.system().lower() == 'windows':
        return '\\'
    else:
        return '/'


def raw_to_function_device_model(message: str) -> FunctionDeviceModel:
    return FunctionDeviceModel(is_raw=True, acoustics='', data=message)


def load_models(path: str):
    importlib.import_module(name=path)


def get_root_path():
    return os.path.dirname(os.path.abspath(__file__))


def get_all_core_packages():
    """
    过于恶心的逻辑，以后优化
    :return:
    """
    g_dirs = list()

    def inner_recursive(path):
        dirs = os.listdir(path)
        for file in dirs:
            new_file = os.path.join(path, file)
            if os.path.isdir(new_file):
                inner_recursive(new_file)
            else:
                g_dirs.append(new_file)
        return g_dirs

    file_abs_paths = inner_recursive(os.path.dirname(os.path.abspath(__file__)) + get_slash() + 'core')
    result = [file_abs_path[:-3]
              for file_abs_path in file_abs_paths
              if file_abs_path.endswith('.py') and not file_abs_path.endswith('__init__.py')]
    return ['iot' + '.'.join(res.split('iot')[-1].split(get_slash())) for res in result]


def process_bar(num, total, text):
    rate = float(num) / total
    ratenum = int(100 * rate)
    r = '\r' + text + '[{}{}]{}%'.format('*' * ratenum, ' ' * (100 - ratenum), ratenum)
    sys.stdout.write(r)
    sys.stdout.flush()


def mkdir(path: str, dirs: Union[str, Iterable]):
    if isinstance(dirs, str):
        dirs = [dirs]
    for dir in dirs:
        if not os.path.exists(path + get_slash() + dir):
            os.makedirs(path + get_slash() + dir)



from typing import Optional

from iot.semantics import SemanticsGroup
from iot.models import SemanticsModel, SemanticsFromEnum, SemanticsRedirectEnum, SemanticsFunc, UniverseNoticeModel


def button_semantics_model_func(*args) -> UniverseNoticeModel:
    return UniverseNoticeModel(
        smt_code='default05',
        topic='self/button',
        device='button',
        verbose=False,
        message="有人按下了按钮"
    )


@SemanticsGroup.add_model
class ButtonSemanticsModel(SemanticsModel):
    code = "default05"
    frm = SemanticsFromEnum.DEVICE
    topic = 'self/button'
    regex: Optional[str] = None
    regex_num: Optional[str] = None
    redirect = SemanticsRedirectEnum.ACOUSTICS
    func: SemanticsFunc = button_semantics_model_func



from iot.factories import PlayerFactory
from iot.integration.audio import AudioClient
from iot.models import Player


@PlayerFactory.set
class BuiltinPlayer(Player):

    def play(self, path: str) -> None:
        AudioClient.play(path)


from concurrent.futures.thread import ThreadPoolExecutor
from typing import Union, Iterable, Optional

from iot.context import Context

from iot.pipeline import Pipeline
from iot.utils import load_models


class Terminal:

    @classmethod
    def load_models(cls, path: Optional[Union[str, Iterable[str]]]):
        if not path:
            return
        if type(path) is Iterable:
            for p in path:
                load_models(p)
        else:
            load_models(path)

    @classmethod
    def run(cls, *args, **kwargs):
        Pipeline.before(**kwargs)
        terminal_mode = Context.config['TERMINAL_MODE']
        if terminal_mode == '0':
            Pipeline.user_loop()
        elif terminal_mode == '1':
            Pipeline.device_loop()
        elif terminal_mode == '2':
            with ThreadPoolExecutor(max_workers=2) as pool:
                pool.submit(Pipeline.user_loop)
                pool.submit(Pipeline.device_loop)
        Pipeline.after()



import json
from typing import ClassVar

from iot.integration.mqtt import MQTTClient, MQTT_Callback
from iot.models import FunctionDeviceModel


class MessagesProcessor:
    __client: ClassVar["MQTTClient"] = None

    def __init__(self, hostname: str, port: int):
        self.__client = MQTTClient(hostname=hostname, port=port)

    def publish(self, model: FunctionDeviceModel):
        data = model.data if model.is_raw else json.dumps(model.data)
        self.__client.publish(topic=model.topic, payload=data)

    def subscribe(self, topics: str, callback: MQTT_Callback):
        self.__client.subscribe(topics=topics, callback=callback)

# mqttx-cli-win-x64.exe sub -t default/light  -p 1883
# mqttx-cli-win-x64.exe pub -t self/button -p 1883 -m {\"topic\":\"self/button\",\"device\":\"button\",\"verbose\":\"false\",\"message\":\"有人按下了按钮\"}



from typing import Callable, Any

from paho.mqtt import publish, subscribe
from paho.mqtt.client import Client, MQTTMessage

MQTT_UserData = Any

MQTT_Client = Client

MQTT_Message = MQTTMessage

MQTT_Callback = Callable[[MQTT_Client, MQTT_UserData, MQTT_Message], None]


class MQTTClient:

    def __init__(self, hostname: str = '127.0.0.1', port: int = 1883):
        self.hostname = hostname
        self.port = port

    def publish(self, topic: str, payload: str):
        publish.single(topic=topic, payload=payload, hostname=self.hostname, port=self.port)

    def subscribe(self, topics: str, callback: MQTT_Callback):
        subscribe.callback(topics=topics, callback=callback, hostname=self.hostname, port=self.port)


from threading import Lock
from typing import ClassVar

from iot.config import Config
from iot.factories import RecorderFactory, PlayerFactory, AsrProcessorFactory, TtsProcessorFactory
from iot.messages import MessagesProcessor
from iot.models import Recorder, Player, AsrProcessor, TtsProcessor
from iot.utils import get_all_core_packages, load_models, mkdir


class Context:
    """
    Context上下文，所有类成员都是单例
    """
    config: ClassVar["Config"] = None
    recorder: ClassVar["Recorder"] = None
    player: ClassVar["Player"] = None
    asr_processor: ClassVar["AsrProcessor"] = None
    tts_processor: ClassVar["TtsProcessor"] = None
    messages_processor: ClassVar["MessagesProcessor"] = None

    uni_lock: Lock = None

    @classmethod
    def load_context(cls):
        if not cls.config:
            cls.config = Config()
            cls.config.load_config()
        cls.__build_context()
        cls.__load_builtins()
        if not cls.recorder:
            cls.recorder = RecorderFactory.get(cls.config['DEVICE_REC'])
        if not cls.player:
            cls.player = PlayerFactory.get(cls.config['DEVICE_PLY'])
        if not cls.asr_processor:
            cls.asr_processor = AsrProcessorFactory.get(cls.config['ASR_PROCESSOR'])
        if not cls.tts_processor:
            cls.tts_processor = TtsProcessorFactory.get(cls.config['TTS_PROCESSOR'])
        if not cls.messages_processor:
            cls.messages_processor = \
                MessagesProcessor(hostname=cls.config['MQTT_HOST'], port=int(cls.config['MQTT_PORT']))
        if not cls.uni_lock:
            cls.uni_lock = Lock()

    @classmethod
    def __load_builtins(cls):
        models = get_all_core_packages()
        for model in models:
            load_models(model)

    @classmethod
    def __build_context(cls):
        mkdir(cls.config['RESOURCE_PATH'], ['wav', 'txt'])



from setuptools import setup


with open("README.md", "r", encoding='UTF-8') as f:
    long_description = f.read()


setup(
    name='fubuki-iot',
    version='0.2',
    python_requires=">=3.8",
    packages=['iot', 'iot.core', 'iot.core.hardware', 'iot.core.acoustics', 'iot.core.semantics', 'iot.integration'],
    install_requires=['keyboard>=0.13.5', 'loguru>=0.6.0', 'paho-mqtt>=1.6.1', 'PyAudio>=0.2.12', 'pydantic>=1.10.2',
                      'python-dotenv>=0.21.0', 'requests>=2.28.1'],
    url='https://github.com/littlebutt/fubuki-iot',
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='MIT License',
    author='littlebutt',
    author_email='luogan1996@icloud.com',
    description='An accessable Iot Terminal implementated by Python'
)



from iot.semantics import SemanticsGroup
from iot.models import SemanticsModel, SemanticsFromEnum, SemanticsRedirectEnum, FunctionDeviceModel, SemanticsFunc


def light_semantics_model_func(*args) -> FunctionDeviceModel:
    param = args[1]
    if param == "卧室":
        return FunctionDeviceModel(smt_code="default03",
                                   topic="light",
                                   is_raw=False,
                                   acoustics="好的，正在为你打开卧室灯",
                                   data={
                                       'position': 'bedroom'
                                   })
    elif param == "客厅":
        return FunctionDeviceModel(smt_code="default03",
                                   topic="light",
                                   is_raw=False,
                                   acoustics="好的，正在为你打开客厅灯",
                                   data={
                                       'position': 'livingroom'
                                   })
    else:
        return FunctionDeviceModel(smt_code="default03",
                                   topic="light",
                                   is_raw=False,
                                   acoustics="好的，正在为你打开餐厅灯",
                                   data={
                                       'position': 'dinningroom'
                                   })


@SemanticsGroup.add_model
class SwitchOnSemanticsModel(SemanticsModel):
    code = "default03"
    frm = SemanticsFromEnum.USER
    topic = 'default/light'
    regex = "打开(.*)灯"
    regex_num = 2
    redirect = SemanticsRedirectEnum.MESSAGE
    func: SemanticsFunc = light_semantics_model_func


import wave
from typing import List

import pyaudio
from loguru import logger

from iot.utils import process_bar


class AudioClient:

    @staticmethod
    def record(format=pyaudio.paInt16,
               channels: int = 1,
               rate: int = 16000,
               input: bool = True,
               frames_per_buffer: int = 1024,
               time: int = 5) -> List[bytes]:
        audio = pyaudio.PyAudio()
        ret_value: List[bytes] = []
        logger.info("Recording starts")
        stream = audio.open(format=format,
                            channels=channels,
                            rate=rate,
                            input=input,
                            frames_per_buffer=frames_per_buffer)
        for i in range(0, int(rate / frames_per_buffer * time)):
            process_bar(i, int(rate / frames_per_buffer * time), "⚪ Recording")
            data: bytes = stream.read(frames_per_buffer)
            ret_value.append(data)
        print('\n')
        logger.info("Recording ends")
        stream.stop_stream()
        audio.terminate()
        return ret_value

    @staticmethod
    def save(data: List[bytes],
             path: str,
             format=pyaudio.paInt16,
             channels: int = 1,
             rate: int = 16000) -> None:
        audio = pyaudio.PyAudio()
        with wave.open(path, 'wb') as wf:
            wf.setnchannels(channels)
            wf.setsampwidth(audio.get_sample_size(format))
            wf.setframerate(rate)
            wf.writeframes(b''.join(data))
            logger.info("wave saved")

    @staticmethod
    def play(path: str, frames_per_buffer: int = 1024):
        with wave.open(path, 'rb') as wf:
            audio = pyaudio.PyAudio()
            stream = audio.open(format=audio.get_format_from_width(wf.getsampwidth()),
                                channels=wf.getnchannels(),
                                rate=wf.getframerate(),
                                output=True)
            data = wf.readframes(frames_per_buffer)
            logger.info("playing starts")
            while data:
                stream.write(data)
                data = wf.readframes(frames_per_buffer)
            stream.stop_stream()
            stream.close()
            logger.info("playing ends")
            audio.terminate()




from iot.semantics import SemanticsGroup
from iot.models import SemanticsModel, SemanticsFromEnum, SemanticsRedirectEnum, SemanticsFunc, FunctionDeviceModel


def dialog_semantics_model_func(*args) -> FunctionDeviceModel:
    return FunctionDeviceModel(
        smt_code='default04',
        is_raw=True,
        acoustics="在的",
        data="在的"
    )


@SemanticsGroup.add_model
class DialogSemanticsModel(SemanticsModel):
    code = "default04"
    frm = SemanticsFromEnum.USER
    topic = ''
    regex = "在吗"
    regex_num = 1
    redirect = SemanticsRedirectEnum.ACOUSTICS
    func: SemanticsFunc = dialog_semantics_model_func


def complex_dialog_semantics_model_func(*args) -> FunctionDeviceModel:
    return FunctionDeviceModel(
        smt_code='default04',
        is_raw=False,
        acoustics="",
        data=""
    )


@SemanticsGroup.add_model
class ComplexDialogSemanticsModel(SemanticsModel):
    code = 'default06'
    frm = SemanticsFromEnum.USER
    topic = ''
    regex = "你好"
    regex_num = 1
    redirect = SemanticsRedirectEnum.SEMANTICS
    func: SemanticsFunc = complex_dialog_semantics_model_func


from loguru import logger

from iot.config import Config
from iot.context import Context
from iot.hooks import HooksGroup
from iot.integration.mqtt import MQTT_UserData, MQTT_Client, MQTT_Message
from iot.semantics import SemanticsGroup
from iot.models import SemanticsRedirectEnum, UniverseNoticeModel, FunctionDeviceModel
from iot.utils import raw_to_function_device_model


class Pipeline:

    @staticmethod
    def before(**kwargs):
        if not kwargs:
            config = Config()
            config.data = kwargs
            Context.config = config
        logger.info("开始加载智能终端上下文")
        Context.load_context()
        logger.info("上下文加载完成 " + Context.__dict__.__str__())
        HooksGroup.execute_start_up(context=Context, semantics_group=SemanticsGroup)

    @staticmethod
    def user_loop():

        while True:
            logger.info("开始监听用户指令")
            if Context.recorder.awake():
                with Context.uni_lock:
                    logger.info("智能终端被唤醒")
                    path = Context.tts_processor.tts("我在，你说")
                    Context.player.play(path)
                    time = int(Context.config['RECORDER_TIME']) if 'RECORDER_TIME' in Context.config else 3
                    path = Context.recorder.record(time)
                    logger.info("已完成录音，路径：" + path)
                    message = Context.asr_processor.asr(path)
                    if message is None:
                        logger.warning("语音识别发生错误")
                        Pipeline.play_error_reminder()
                        continue
                    logger.info("已完成语音识别，内容：" + message)
                    model = raw_to_function_device_model(message)
                    HooksGroup.execute_model_preprocess(context=Context, model=model)
                    result, redirection = SemanticsGroup.process_usr_command(model)
                    HooksGroup.execute_model_postprocessor(context=Context, model=model)
                    if redirection == SemanticsRedirectEnum.ACOUSTICS:
                        logger.info("语义转换成功，重定向至ACOUSTICS")
                        path = Context.tts_processor.tts(result.acoustics if result.acoustics else result.data)
                        Context.player.play(path)
                        continue
                    elif redirection == SemanticsRedirectEnum.MESSAGE:
                        logger.info("语义转换成功，重定向至MESSAGE")
                        Context.messages_processor.publish(result)
                        path = Context.tts_processor.tts(result.acoustics)
                        Context.player.play(path)
                        continue
                    else:
                        logger.info("语义转换失败")
                        Pipeline.play_error_reminder()
                        continue
            break

    @staticmethod
    def device_loop():
        def inner_callback(client: MQTT_Client, user_data: MQTT_UserData, message: MQTT_Message):
            with Context.uni_lock:
                payload = str(message.payload, 'utf-8')
                logger.info(f"获取到来自设备消息，topic：{message.topic} payload：{payload}")
                universe_notice_model = None
                try:
                    universe_notice_model = UniverseNoticeModel.parse_raw(payload)
                    logger.success("成功转换成统一推送模型")
                except Exception as e:
                    logger.error("转换成统一推送模型失败 " + e.__str__())
                    return
                universe_notice_model.topic = message.topic
                HooksGroup.execute_model_preprocess(context=Context, model=universe_notice_model)
                result, redirection = SemanticsGroup.process_dvc_command(universe_notice_model)
                HooksGroup.execute_model_postprocessor(context=Context, model=universe_notice_model)
                if not result.verbose:
                    logger.info("语义转换成功，重定向至ACOUSTICS并直接返回")
                    path = Context.tts_processor.tts(result.message)
                    Context.player.play(path)
                else:
                    if redirection == SemanticsRedirectEnum.MESSAGE:
                        logger.info("语义转换成功，重定向至MESSAGE")
                        model = FunctionDeviceModel.parse_obj(result.data)
                        Context.messages_processor.publish(model)
                    elif redirection == SemanticsRedirectEnum.ACOUSTICS:
                        logger.info("语义转换成功，重定向至ACOUSTICS")
                        path = Context.tts_processor.tts(result.message)
                        Context.player.play(path)
                    else:
                        logger.info("语义转换失败")
                        Pipeline.play_error_reminder()

        logger.info("开始监听设备消息...")
        Context.messages_processor.subscribe('self/#', inner_callback)

    @staticmethod
    def after():
        HooksGroup.execute_tear_down(context=Context, semantics_group=SemanticsGroup)

    @staticmethod
    def play_error_reminder() -> None:
        reminder = "抱歉，好像遇到了点问题"
        path = Context.tts_processor.tts(reminder)
        Context.player.play(path)


from typing import Dict, ClassVar

from iot.models import Factory, Recorder, Player, AsrProcessor, TtsProcessor


class RecorderFactory(Factory):
    __recorder: ClassVar[Dict[str, "Recorder"]] = {}

    @classmethod
    def set(cls, target):
        cls.__recorder[target.__name__] = target()

    @classmethod
    def get(cls, target_name):
        if target_name not in cls.__recorder:
            raise KeyError("Cannot find Recorder")
        return cls.__recorder[target_name]


class PlayerFactory(Factory):
    __player: ClassVar[Dict[str, "Player"]] = {}

    @classmethod
    def set(cls, target):
        cls.__player[target.__name__] = target()

    @classmethod
    def get(cls, target_name):
        if target_name not in cls.__player:
            raise KeyError("Cannot find Player")
        return cls.__player[target_name]


class AsrProcessorFactory(Factory):
    __asr_processor: ClassVar[Dict[str, "AsrProcessor"]] = {}

    @classmethod
    def set(cls, target):
        cls.__asr_processor[target.__name__] = target()

    @classmethod
    def get(cls, target_name):
        if target_name not in cls.__asr_processor:
            raise KeyError("Cannot find AsrProcesor")
        return cls.__asr_processor[target_name]


class TtsProcessorFactory(Factory):
    __tts_processor: ClassVar[Dict[str, "TtsProcessor"]] = {}

    @classmethod
    def set(cls, target):
        cls.__tts_processor[target.__name__] = target()

    @classmethod
    def get(cls, target_name):
        if target_name not in cls.__tts_processor:
            raise KeyError("Cannot find TtsProcessor")
        return cls.__tts_processor[target_name]